import math
import numpy as np
import os
import re
import glob
import random

from helper_code import *
import pandas as pd
from sklearn import preprocessing

import tensorflow as tf
import tensorflow_addons as tfa
from tensorflow.keras.models import model_from_json

import heart_sound_preprocess as hsp


# Load patient data as a dict.
def load_patient_data2dict(filename):
    dic = []
    with open(filename, 'r') as f:
        for idx, line in enumerate(f.readlines()):
            if idx == 0:
                line = line.rstrip('\n')
                line = line.split(' ')
                value = np.array(line, np.int16)
            elif idx > value[1]:
                line = line.rstrip('\n')
                line = line.strip('#')
                item = line.split(': ')
                dic.append(item)
    dic = dict(dic)
    return dic

def get_murmur_locations(filename):
    pid_info = load_patient_data2dict(filename)
    
    return pid_info['Murmur locations']

def load_label(label_path):
    df = pd.read_csv(label_path, names=['START', 'END', 'STATE'], sep='\t')
    starts = np.array(df['START']) * 4000
    starts = starts.astype(int)
    states = np.array(df['STATE'])

    return starts, states

def load_label_physionet2022(label_path, fs):
    df = pd.read_csv(label_path, names=['START', 'END', 'STATE'], sep='\t')
    starts = np.array(df['START']) * fs
    starts = starts.astype(int)
    states = np.array(df['STATE'])

    return starts, states

def split_label_physionet2022(label_path, fs):
    starts, states = load_label_physionet2022(label_path, fs)
    s1 = []
    s2 = []
    sys = []
    dia = []
    noise = []
    count = 0
    for (start, state) in zip(starts, states):
        index = np.array([start, starts[count+1]]) if count < len(starts)-1 else np.array([start])
        if state == 1:
            s1.append(index)
        elif state == 3:
            s2.append(index)
        elif state == 2:
            sys.append(index)
        elif state == 4:
            dia.append(index)
        elif state == 0:
            noise.append(index)

        count += 1

    return s1, s2, sys, dia, noise

def gen_label_physionet2022(label_path_, wav_len, fs):
    s1, s2, sys, dia, noise = split_label_physionet2022(label_path_, fs)
    labels = np.zeros((wav_len, ), dtype=np.int16)
    # s1+sys+s2=1; dia=0, noise=2
    for duration in s1:
        end = duration[1] if len(duration) == 2 else -1
        labels[duration[0]: end] = 1
    for duration in s2:
        end = duration[1] if len(duration) == 2 else -1
        labels[duration[0]: end] = 1
    for duration in sys:
        end = duration[1] if len(duration) == 2 else -1
        labels[duration[0]: end] = 1
    for duration in dia:
        end = duration[1] if len(duration) == 2 else -1
        labels[duration[0]: end] = 0
    for duration in noise:
        end = duration[1] if len(duration) == 2 else -1
        labels[duration[0]: end] = 2

    return labels

def extract_no_annotation_noise(data, label):
    noise_index = np.where(label == 2)[0]
    noise = data[noise_index]
    
    return noise

def ngrams(data, length, fs, overlap):
    grams = []
    for i in range(0, length-fs*5, fs*overlap):
        grams.append(data[i: i+fs*5])
        
    return grams

def en_est(logits):
    logits_p = np.expand_dims(logits[:, :, 1], -1)
    logits_n = np.expand_dims(logits[:, :, 2], -1)
    logits_p_s = logits_p / (logits_p - logits_n + 1.)
    logits_n_s = (1. - logits_n) / (logits_p - logits_n + 1.)
    
    p = np.concatenate((logits_p_s, logits_n_s), -1)
    
    logp = np.log(p + 1e-5)
    plogp = p * logp
    entropy = -np.sum(plogp, axis=-1)
    entropy = np.mean(entropy, axis=-1)
    
    return entropy

def mi_est(logits):
    logits_o = logits[:, :, 0]
    logits_p = logits[:, :, 1]
    logits_n = logits[:, :, 2]
    eu_p = logits_p * np.log(logits_p/(logits_o + 1e-5) + 1e-5)
    eu_n = (1. - logits_n) * np.log((1. - logits_n)/(1. - logits_o + 1e-5) + 1e-5)

    eu = eu_p + eu_n
    eu = np.mean(eu, axis=-1)
  
    return eu

def gen_training_data(data_path):

    normal_save_path = './murmur_training_set/normal_data/'
    murmur_save_path = './murmur_training_set/murmur_data/'
    unknown_save_path = './murmur_training_set/unknown_data/'

    if not os.path.exists(normal_save_path):
        os.makedirs(normal_save_path)
        
    if not os.path.exists(murmur_save_path):
        os.makedirs(murmur_save_path)
        
    if not os.path.exists(unknown_save_path):
        os.makedirs(unknown_save_path)

    with open('./seg_model/hs_segmentor_ua.json', 'r') as json_file:
        config= json_file.read()
    seg_model = model_from_json(config, custom_objects={'InstanceNormalization':tfa.layers.InstanceNormalization, 'leaky_relu': tf.nn.leaky_relu})
    seg_model.load_weights('./seg_model/hs_segmentor_ua.h5')

    def beat_segment4id(data, sr, overlap):
        
        hs_batch = ngrams(data, len(data), sr, overlap)
        hs_batch = np.array(hs_batch)
        hs_batch_pp = hsp.hs_process_batch4seg(hs_batch, sr)

        logits_batch = seg_model(hs_batch_pp, training=False)
        eu_batch = mi_est(logits_batch) # (B, )
        
        index_id_quality = np.where(eu_batch <= 0.3)[0]
        hs_id_quality = hs_batch[index_id_quality]
        
        return hs_id_quality

    def is_tsv(l):
        return l.endswith('.tsv')
    
    physionet2022_ref_files = list(filter(is_tsv, os.listdir(data_path)))
    patient_files = find_patient_files(data_path)

    ### Add non-annoated noise into unknown data
    noise_count = 0
    for ref_file in physionet2022_ref_files:
        index = re.split('[.]', ref_file)[0]
        print(index)
        record = index + '.wav'
        
        ref_path = os.path.join(data_path, ref_file)
        hs_path = os.path.join(data_path, record)
        
        if not os.path.exists(hs_path):
            continue
        
        data, sr = load_wav_file(hs_path)
        label = gen_label_physionet2022(ref_path, len(data), sr)
        
        noise_data = extract_no_annotation_noise(data, label)
        
        if len(noise_data) >= 20000:
            id_hss = beat_segment4id(noise_data, sr, overlap=3)
            
            for i, id_hs in enumerate(id_hss):
                noise_id_hs = hsp.hs_process4murmur(id_hs, sr)
                np.save(os.path.join(unknown_save_path, 'noise_{}.npy'.format(str(noise_count))), noise_id_hs)
                noise_count += 1

    for pid_file in patient_files:
        pid_data = load_patient_data(pid_file)
        pid = get_patient_id(pid_data)
        p_locs = get_locations(pid_data)
        murmur_flag = get_murmur(pid_data)
        
        if murmur_flag == 'Present':
            p_locs = get_murmur_locations(pid_file)
            if p_locs != 'nan':
                p_locs = p_locs.split('+')
            else:
                p_locs = []
            p_locs = np.array(p_locs)
            p_locs_unique, counts = np.unique(p_locs, return_counts=True)
        
        else:
            p_locs = np.array(p_locs)
            p_locs_unique, counts = np.unique(p_locs, return_counts=True)
            
        for i, loc in enumerate(p_locs_unique):
            
            target_wav_files = glob.glob(os.path.join(data_path, '{}_{}*.wav'.format(pid, loc)))
            
            for wav_path in target_wav_files:

                wav_data, sr = load_wav_file(wav_path)

                if murmur_flag == 'Present':
                    id_hss = beat_segment4id(wav_data, sr, overlap=1)

                    for j, id_hs in enumerate(id_hss):
                        id_hs_pp = hsp.hs_process4murmur(id_hs, sr)
                        np.save(os.path.join(murmur_save_path, '{}_{}_{}.npy'.format(pid, loc, str(j))), id_hs_pp)

                elif murmur_flag == 'Absent':
                    id_hss = beat_segment4id(wav_data, sr, overlap=5)

                    for j, id_hs in enumerate(id_hss):
                        id_hs_pp = hsp.hs_process4murmur(id_hs, sr)
                        np.save(os.path.join(normal_save_path, '{}_{}_{}.npy'.format(pid, loc, str(j))), id_hs_pp)

                elif murmur_flag == 'Unknown':
                    id_hss = beat_segment4id(wav_data, sr, overlap=1)

                    for j, id_hs in enumerate(id_hss):
                        id_hs_pp = hsp.hs_process4murmur(id_hs, sr)
                        np.save(os.path.join(unknown_save_path, '{}_{}_{}.npy'.format(pid, loc, str(j))), id_hs_pp)
    
    def is_npy(l):
        return l.endswith('.npy')

    def del_npy(f):
        ind = re.split('[.]', f)[0]
        return ind
    
    normal_files = os.listdir(normal_save_path)
    murmur_files = os.listdir(murmur_save_path)
    unknown_files = os.listdir(unknown_save_path)

    with open('./murmur_training_set/NORMAL_RECORDS', 'w') as record_list:
        for index in normal_files:
            print(index, file=record_list)

        record_list.close()

    with open('./murmur_training_set/MURMUR_RECORDS', 'w') as record_list:
        for index in murmur_files:
            print(index, file=record_list)

        record_list.close()

    with open('./murmur_training_set/UNKNOWN_RECORDS', 'w') as record_list:
        for index in unknown_files:
            print(index, file=record_list)

        record_list.close()