#!/usr/bin/env python

# Edit this script to add your team's code. Some functions are *required*, but you can edit most parts of the required functions,
# change or remove non-required functions, and add your own functions.

################################################################################
#
# Import libraries and functions. You can change or remove them.
#
################################################################################

import os

from helper_code import *
import numpy as np, scipy as sp, scipy.stats, os, sys, joblib
from sklearn.impute import SimpleImputer
from sklearn.ensemble import RandomForestClassifier

import tensorflow as tf
import tensorflow_addons as tfa
from tensorflow.keras.models import model_from_json

import heart_sound_preprocess as hsp
from train import HSTrainer
from module.architecture import *
from module.prepare_dataset import HSDataset
from module.utils import *
from plus_functions import *

################################################################################
#
# Required functions. Edit these functions to add your code, but do not change the arguments.
#
################################################################################

# Train your model.
def train_challenge_model(data_folder, model_folder, verbose):

    # Find data files.
    if verbose >= 1:
        print('Finding data files...')
    
    # Find the patient data files.
    patient_files = find_patient_files(data_folder)
    num_patient_files = len(patient_files)

    if num_patient_files==0:
        raise Exception('No data was provided.')

    # Create a folder for the model if it does not already exist.
    os.makedirs(model_folder, exist_ok=True)

    classes = ['Absent', 'Present', 'Unknown']
    num_classes = len(classes)

    # Extract the features and labels.
    if verbose >= 1:
        print('Extracting features and labels from the Challenge data...')

    labels = list()

    ###### Prepare training beats #########
    print("Preparing Training Episodes......")

    gen_training_data(data_folder)

    print("Complete!")
    ###### Prepare training beats #########
    
    ###### Train the model ################
    print("Start Training!")

    if verbose >= 1:
        print('Training model...')
    
    DATA_PATH = './murmur_training_set/'
    NORMAL_INDEX = './murmur_training_set/NORMAL_RECORDS'
    MURMUR_INDEX = './murmur_training_set/MURMUR_RECORDS'
    UNKNOWN_INDEX = './murmur_training_set/UNKNOWN_RECORDS'
    EPOCHES = 80
    BATCH_SIZE = 100
    LR1 = 0.001
    TRAIN_SIZE = 5000

    if not os.path.exists(model_folder):
        os.makedirs(model_folder)

    dataset = HSDataset(batch_size=BATCH_SIZE, path=DATA_PATH, normal_list=NORMAL_INDEX, murmur_list=MURMUR_INDEX, unknown_list=UNKNOWN_INDEX)

    en = encoder()
    de = decoder()

    hs_trainer = HSTrainer(BATCH_SIZE, en, de, y_loss_func4hs, LR1, opt_func, model_folder)
    hs_trainer.train(BATCH_SIZE, TRAIN_SIZE, EPOCHES, dataset)

    ### Summary
    hs_trainer = HSTrainer(BATCH_SIZE, en, de, y_loss_func4hs, LR1, opt_func, model_folder)
    x_in = Input(shape=(5000, 1))

    z = hs_trainer.en(x_in, training=False)
    logits = hs_trainer.de(z, training=False)
    
    murmur_detector = Model(x_in, logits, name='murmur_detector')
    murmur_detector.summary()
    config = murmur_detector.to_json()
    with open(os.path.join(model_folder, 'murmur_detector.json'), 'w') as json_file:
        json_file.write(config)
    murmur_detector.save_weights(os.path.join(model_folder, 'murmur_detector.h5'))

    if verbose >= 1:
        print('Done.')
    
# Load your trained model. This function is *required*. You should edit this function to add your code, but do *not* change the
# arguments of this function.
def load_challenge_model(model_folder, verbose):
    with open(os.path.join(model_folder, 'murmur_detector.json'), 'r') as json_file:
        config= json_file.read()
    model = model_from_json(config, custom_objects={'InstanceNormalization':tfa.layers.InstanceNormalization, 'leaky_relu': tf.nn.leaky_relu})
    model.load_weights(os.path.join(model_folder, 'murmur_detector.h5'))

    return model

# load the segmentation model.
def load_segment_model(model_folder):
    with open(os.path.join(model_folder, 'hs_segmentor_ua.json'), 'r') as json_file:
        config= json_file.read()
    model = model_from_json(config, custom_objects={'InstanceNormalization':tfa.layers.InstanceNormalization, 'leaky_relu': tf.nn.leaky_relu})
    model.load_weights(os.path.join(model_folder, 'hs_segmentor_ua.h5'))

    return model

# Run your trained model. This function is *required*. You should edit this function to add your code, but do *not* change the
# arguments of this function.
def run_challenge_model(model, data, recordings, verbose):

    seg_model = load_segment_model('./seg_model/')

    classes = ['Absent', 'Present', 'Unknown']
    murmur_flags = [] # 0: normal; 1: murmur; 2: unknown
    probabilities = []
    
    for i in range(len(recordings)):

        recording = recordings[i]
        sr = 4000
        
        if len(recording) <= sr * 5:
            recording = np.concatenate((recording, recording), axis=0)

        hs_batch = ngrams(recording, len(recording), sr, overlap=3)
        hs_batch = np.array(hs_batch)
        hs_batch_pp = hsp.hs_process_batch4seg(hs_batch, sr)

        seg_logits_batch = seg_model(hs_batch_pp, training=False)
        eu_batch = mi_est(seg_logits_batch)

        index_best_quality = np.argmin(eu_batch, axis=0)
        hs_best = hs_batch[index_best_quality]

        hs_best_pp = hsp.hs_process4murmur(hs_best, sr)
        hs_best_pp = np.expand_dims(hs_best_pp, axis=0)

        probability = model(hs_best_pp, training=False) # (1, 3)
        probability = np.squeeze(probability)

        murmur_flag = np.argmax(probability)
        probabilities.append(probability)
        murmur_flags.append(murmur_flag)
    
    probabilities = np.array(probabilities) # (len(recordings), 3)
    labels = np.zeros(len(classes), dtype=np.int16)

    ### decision strategy ###
    if 1 in murmur_flags:
        labels[1] = 1
        idx = np.argmax(probabilities[:, 1], axis=0)
        probabilities_selected = probabilities[idx]
    elif 2 in murmur_flags:
        labels[2] = 1
        idx = np.argmax(probabilities[:, 2], axis=0)
        probabilities_selected = probabilities[idx]
    else:
        labels[0] = 1
        idx = np.argmin(probabilities[:, 0], axis=0)
        probabilities_selected = probabilities[idx]

    return classes, labels, probabilities_selected

################################################################################
#
# Optional functions. You can change or remove these functions and/or add new functions.
#
################################################################################

