import os
import time
from tqdm import tqdm
import numpy as np
import random

import tensorflow as tf
import tensorflow.keras.backend as K
from tensorflow.keras.models import model_from_json, load_model
from tensorflow.keras import layers
tf.keras.backend.set_floatx('float32')
tf.keras.backend.clear_session()
from tensorflow.keras.callbacks import LearningRateScheduler, History
from tensorflow.keras.models import Model
from tensorflow.keras import Input

from module.architecture import *
from module.prepare_dataset import HSDataset
from module.utils import *


class HSTrainer():
    def __init__(self, batch_size, encoder, decoder, y_loss, alpha_lr, optimizer, model_save_path):
        self.now = None
        self.batch_size = batch_size
        self.en = encoder
        self.de = decoder

        self.y_loss = y_loss
        self.alpha_optimizer = optimizer(alpha_lr)
        self.model_save_path = model_save_path   

        self.en_checkpoint = tf.train.Checkpoint(step=tf.Variable(0),
                                                 psnr=tf.Variable(0),
                                                 optimizer=self.alpha_optimizer,
                                                 model=self.en)
        self.en_checkpoint_manager = tf.train.CheckpointManager(checkpoint=self.en_checkpoint,
                                                                directory=os.path.join(self.model_save_path, 'en_ckpt'),
                                                                max_to_keep=3)
        self.de_checkpoint = tf.train.Checkpoint(step=tf.Variable(0),
                                                 psnr=tf.Variable(0),
                                                 optimizer=self.alpha_optimizer,
                                                 model=self.de)
        self.de_checkpoint_manager = tf.train.CheckpointManager(checkpoint=self.de_checkpoint,
                                                                directory=os.path.join(self.model_save_path, 'de_ckpt'),
                                                                max_to_keep=3)

        self.restore()

        self.train_loss_mean = tf.keras.metrics.Mean(name='train_loss')
        self.train_acc = tf.keras.metrics.BinaryAccuracy(name='train_acc')

        # self.val_loss_mean = tf.keras.metrics.Mean(name='test_loss')
        # self.val_acc = tf.keras.metrics.BinaryAccuracy(name='test_acc')

    # Train the model
    def train(self, batch_size, train_size, epoches, dataset):
        en_ckpt_mgr = self.en_checkpoint_manager
        de_ckpt_mgr = self.de_checkpoint_manager

        train_iters = train_size // batch_size

        summary_writer = tf.summary.create_file_writer(os.path.join(self.model_save_path, 'log'))

        # best_acc = 0.0
        # wait = 0
        # patience = 20
        with summary_writer.as_default():
            for epoch in tqdm(range(epoches)):
                for step in range((epoch*train_iters), (epoch+1)*train_iters):
                    batch_x, batch_y = dataset.inputs(is_training=True)
                    
                    loss, logits = self.train_step(batch_x, batch_y)

                    self.train_loss_mean(loss)
                    self.train_acc(batch_y, logits)

                    if step%10 == 0:
                        step_template = 'Epoch {}, Step {}/{}, Loss: {}, Acc: {}'
                        print (step_template.format(epoch+1,
                                                    step-(epoch*train_iters),
                                                    train_iters,
                                                    self.train_loss_mean.result(),
                                                    self.train_acc.result()*100))
                    self.train_loss_mean.reset_states()
                    self.train_acc.reset_states()

                en_ckpt_mgr.save()
                de_ckpt_mgr.save()

                """
                # test data
                val_x, val_y = dataset.inputs(is_training=False)
                _ = self.val_step(val_x, val_y)
                val_template = 'Epoch {}, Test Loss: {}, Test Acc: {}'
                print (val_template.format(epoch+1,
                                           self.val_loss_mean.result(),
                                           self.val_acc.result()*100))
                # tf.summary.scalar("my_metric", 0.5, step=epoch)
                # summary_writer.flush()

                # psnr_value = self.test_acc.result()
                
                val_acc = self.val_acc.result()

                wait += 1
                if val_acc > best_acc:
                    best_acc = val_acc

                    en_ckpt_mgr.save()
                    de_ckpt_mgr.save()

                    wait = 0
                if wait >= patience:
                    
                    break
                    
                self.val_loss_mean.reset_states()
                self.val_acc.reset_states()
                """

        print ('training done.')

    @tf.function
    def train_step(self, x, label):

        with tf.GradientTape() as alpha_tape:
            z = self.en(x, training=True)
            logits = self.de(z, training=True)
            
            loss = self.y_loss(logits, label)

        learnable_variables_alpha = self.en.trainable_variables + self.de.trainable_variables
        
        # Get the gradient
        alpha_gradients = alpha_tape.gradient(loss, learnable_variables_alpha)

        # Update weights
        self.alpha_optimizer.apply_gradients(zip(alpha_gradients, learnable_variables_alpha))

        return loss, logits

    """
    @tf.function
    def val_step(self, x, labels):

        z = self.en(x, training=False)
        val_logits = self.de(z, training=False)

        val_loss = self.y_loss(val_logits, labels)

        self.val_loss_mean(val_loss)
        self.val_acc(labels, val_logits)

        return val_loss
    """

    def restore(self):
        if self.en_checkpoint_manager.latest_checkpoint:
            self.en_checkpoint.restore(self.en_checkpoint_manager.latest_checkpoint)
        if self.de_checkpoint_manager.latest_checkpoint:
            self.de_checkpoint.restore(self.de_checkpoint_manager.latest_checkpoint)


if __name__ == '__main__':
    DATA_PATH = '/data/heart_sound/physionet2022/murmur_training_set/'
    NORMAL_INDEX = '/data/heart_sound/physionet2022/murmur_training_set/NORMAL_RECORDS'
    MURMUR_INDEX = '/data/heart_sound/physionet2022/murmur_training_set/MURMUR_RECORDS'
    UNKNOWN_INDEX = '/data/heart_sound/physionet2022/murmur_training_set/UNKNOWN_RECORDS'
    # NOISE_INDEX = '/data/heart_sound/physionet2022/murmur_training_set/NOISE_RECORDS'
    EPOCHES = 30
    BATCH_SIZE = 100
    LR1 = 0.001
    TRAIN_SIZE = 5000

    MODEL_SAVE_PATH = './murmur_model2/'

    if not os.path.exists(MODEL_SAVE_PATH):
        os.makedirs(MODEL_SAVE_PATH)

    dataset = HSDataset(batch_size=BATCH_SIZE, path=DATA_PATH, normal_list=NORMAL_INDEX, murmur_list=MURMUR_INDEX, unknown_list=UNKNOWN_INDEX)

    en = encoder()
    de = decoder()

    hs_trainer = HSTrainer(BATCH_SIZE, en, de, y_loss_func4hs, LR1, opt_func, MODEL_SAVE_PATH)
    hs_trainer.train(BATCH_SIZE, TRAIN_SIZE, EPOCHES, dataset)

    ### Summary
    hs_trainer = HSTrainer(BATCH_SIZE, en, de, y_loss_func4hs, LR1, opt_func, MODEL_SAVE_PATH)
    x_in = Input(shape=(5000, 1))

    z = hs_trainer.en(x_in, training=False)
    logits = hs_trainer.de(z, training=False)
    
    murmur_detector = Model(x_in, logits, name='murmur_detector')
    murmur_detector.summary()
    config = murmur_detector.to_json()
    with open(os.path.join(MODEL_SAVE_PATH, 'murmur_detector.json'), 'w') as json_file:
        json_file.write(config)
    murmur_detector.save_weights(os.path.join(MODEL_SAVE_PATH, 'murmur_detector.h5'))